"""A Micropub server."""

import pathlib
import random

import pendulum
import vobject
# import webmention
import websub
from microformats import discover_post_type
from understory import sql, web
from understory.web import tx

__all__ = ["get_config"]


class PostNotFoundError(Exception):
    """Post could not be found."""


posts_model = sql.model(
    f"{__name__}.posts",
    resources={
        "permalink": "TEXT UNIQUE",
        "version": "TEXT UNIQUE",
        "resource": "JSON",
    },
    deleted_resources={
        "permalink": "TEXT UNIQUE",
        "version": "TEXT UNIQUE",
        "resource": "JSON",
    },
    syndication={"destination": "JSON NOT NULL"},
)
media_model = sql.model(
    f"{__name__}.media",
    media={"mid": "TEXT", "sha256": "TEXT UNIQUE", "size": "INTEGER"},
)

# TODO supported_types = {"RSVP": ["in-reply-to", "rsvp"]}


def get_config():
    """"""
    syndication_endpoints = []
    # TODO "channels": generate_channels()}
    return {
        "q": ["category", "contact", "source", "syndicate-to"],
        "media-endpoint": f"https://{tx.host.name}/pub/media",
        "syndicate-to": syndication_endpoints,
        "visibility": ["public", "unlisted", "private"],
        "timezone": "America/Los_Angeles",
    }


def generate_trailer():
    letterspace = "abcdefghijkmnopqrstuvwxyz23456789"
    trailer = "".join([random.choice(letterspace) for i in range(2)])
    if trailer in ("bs", "ok", "hi", "oz", "lb"):
        return generate_trailer()
    else:
        return trailer


@posts_model.control
def create(db, resource_type, **resource):
    """Create a resource."""
    for k, v in resource.items():
        if not isinstance(v, list):
            resource[k] = [v]
        flat_values = []
        for v in resource[k]:
            if isinstance(v, dict):
                if not ("html" in v or "datetime" in v):
                    v = dict(**v["properties"], type=[v["type"][0].removeprefix("h-")])
            flat_values.append(v)
        resource[k] = flat_values

    config = get_config()
    # TODO deal with `updated`/`drafted`?
    if "published" in resource:
        # TODO accept simple eg. published=2020-2-20, published=2020-2-20T02:22:22
        # XXX resource["published"][0]["datetime"] = pendulum.from_format(
        # XXX     resource["published"][0]["datetime"], "YYYY-MM-DDTHH:mm:ssZ"
        # XXX )
        # XXX published = resource["published"]
        pass
    else:
        resource["published"] = [
            {
                "datetime": web.utcnow().isoformat(),
                "timezone": config["timezone"],
            }
        ]
    published = pendulum.parse(
        resource["published"][0]["datetime"],
        tz=resource["published"][0]["timezone"],
    )

    resource["visibility"] = resource.get("visibility", ["public"])
    # XXX resource["channel"] = resource.get("channel", [])
    mentions = []
    urls = resource.pop("url", [])
    if resource_type == "card":
        slug = resource.get("nickname", resource.get("name"))[0]
        urls.insert(0, f"/pub/cards/{web.textslug(slug)}")
    elif resource_type == "feed":
        name_slug = web.textslug(resource["name"][0])
        try:
            slug = resource["slug"][0]
        except KeyError:
            slug = name_slug
        resource.update(uid=[slug if slug else name_slug])
        resource.pop("channel", None)
        # XXX urls.insert(0, f"/{slug}")
    elif resource_type == "entry":
        #                                         REQUEST URL
        # 1) given: url=/xyz                        => look for exact match
        #     then: url=[/xyz, /2021/3/5...]
        # 2) given: channel=abc, slug=foo           => construct
        #     then: url=[/2021/3/5...]
        # 3) given: no slug                         => only via permalink
        #     then: url=[/2021/3/5...]
        post_type = discover_post_type(resource)
        slug = None
        if post_type == "article":
            slug = resource["name"][0]
        elif post_type == "bookmark":
            mentions.append(resource["bookmark-of"][0])
        elif post_type == "like":
            mentions.append(resource["like-of"][0])
        elif post_type == "rsvp":
            mentions.append(resource["in-reply-to"][0])
        # elif post_type == "identification":
        #     identifications = resource["identification-of"]
        #     identifications[0] = {"type": "cite",
        #                           **identifications[0]["properties"]}
        #     textslug = identifications[0]["name"]
        #     mentions.append(identifications[0]["url"])
        # elif post_type == "follow":
        #     follows = resource["follow-of"]
        #     follows[0] = {"type": "cite", **follows[0]["properties"]}
        #     textslug = follows[0]["name"]
        #     mentions.append(follows[0]["url"])
        #     tx.sub.follow(follows[0]["url"])
        # TODO user indieauth.server.get_identity() ??
        # XXX author_id = list(db.select("identities"))[0]["card"]
        # XXX author_id = get_card()db.select("resources")[0]["card"]["version"]
        resource.update(author=[tx.origin])
    # elif resource_type == "event":
    #     slug = resource.get("nickname", resource.get("name"))[0]
    #     urls.insert(0, f"/pub/cards/{web.textslug(slug)}")
    #     # if resource["uid"] == str(web.uri(tx.host.name)):
    #     #     pass
    resource.update(url=urls, type=[resource_type])
    permalink_base = f"/{web.timeslug(published)}"

    while True:
        permalink = f"{permalink_base}/{generate_trailer()}"
        resource["url"].append(permalink)
        try:
            db.insert(
                "resources",
                permalink=permalink,
                version=web.nbrandom(10),
                resource=resource,
            )
        except db.IntegrityError:
            continue
        break
    return permalink, mentions


@posts_model.control
def read(db, url):
    """Return an entry with its metadata."""
    if not url.startswith(("http://", "https://")):
        url = f"/{url.strip('/')}"
    try:
        resource = db.select(
            "resources",
            where="""json_extract(resources.resource, '$.url[0]') == ?""",
            vals=[url],
        )[0]
    except IndexError:
        resource = db.select(
            "resources",
            where="""json_extract(resources.resource, '$.alias[0]') == ?""",
            vals=[url],
        )[0]
    r = resource["resource"]
    if "entry" in r["type"]:
        r["author"] = tx.identities.get_identity(r["author"][0])["card"]
    return resource


@posts_model.control
def update(db, url, add=None, replace=None, remove=None):
    """Update a resource."""
    permalink = f"/{url.strip('/')}"
    resource = db.select("resources", where="permalink = ?", vals=[permalink])[0][
        "resource"
    ]
    if add:
        for prop, vals in add.items():
            try:
                resource[prop].extend(vals)
            except KeyError:
                resource[prop] = vals
    if replace:
        for prop, vals in replace.items():
            resource[prop] = vals
    if remove:
        for prop, vals in remove.items():
            del resource[prop]
    resource["updated"] = web.utcnow()
    db.update("resources", resource=resource, where="permalink = ?", vals=[permalink])
    # TODO web.publish(url, f".{prop}[-0:-0]", vals)


@posts_model.control
def delete(db, url):
    """Delete a resource."""
    resource = tx.pub.read(url)
    with db.transaction as cur:
        cur.insert("deleted_resources", **resource)
        cur.delete("resources", where="permalink = ?", vals=[url])


@posts_model.control
def search(db, query):
    """Return a list of resources containing `query`."""
    where = """json_extract(resources.resource,
                   '$.bookmark-of[0].url') == ?
               OR json_extract(resources.resource,
                   '$.like-of[0].url') == ?"""
    return db.select("resources", vals=[query, query], where=where)


@posts_model.control
def get_identity(db, version):
    """Return a snapshot of an identity at given version."""
    return tx.pub.get_version(version)


@posts_model.control
def get_version(db, version):
    """Return a snapshot of resource at given version."""
    return db.select("resources", where="version = ?", vals=[version])[0]


@posts_model.control
def get_entry(db, path):
    """"""


@posts_model.control
def get_card(db, nickname):
    """Return the card with given nickname."""
    resource = db.select(
        "resources",
        vals=[nickname],
        where="""json_extract(resources.resource,
                                         '$.nickname[0]') == ?""",
    )[0]
    return resource["resource"]


@posts_model.control
def get_event(db, path):
    """"""


@posts_model.control
def get_entries(db, limit=20, modified="DESC"):
    """Return a list of entries."""
    return db.select(
        "resources",
        order=f"""json_extract(resources.resource,
                                      '$.published[0]') {modified}""",
        where="""json_extract(resources.resource,
                                     '$.type[0]') == 'entry'""",
        limit=limit,
    )


@posts_model.control
def get_cards(db, limit=20):
    """Return a list of alphabetical cards."""
    return db.select(
        "resources",  # order="modified DESC",
        where="""json_extract(resources.resource,
                                     '$.type[0]') == 'card'""",
    )


@posts_model.control
def get_rooms(db, limit=20):
    """Return a list of alphabetical rooms."""
    return db.select(
        "resources",  # order="modified DESC",
        where="""json_extract(resources.resource,
                                     '$.type[0]') == 'room'""",
    )


@posts_model.control
def get_channels(db):
    """Return a list of alphabetical channels."""
    return db.select(
        "resources",  # order="modified DESC",
        where="""json_extract(resources.resource,
                                     '$.type[0]') == 'feed'""",
    )


@posts_model.control
def get_categories(db):
    """Return a list of categories."""
    return [
        r["value"]
        for r in db.select(
            "resources, json_each(resources.resource, '$.category')",
            what="DISTINCT value",
        )
    ]


@posts_model.control
def get_posts(db):
    """."""
    for post in db.select(
        "resources",
        where="""json_extract(resources.resource, '$.channel[0]') IS NULL AND
                 json_extract(resources.resource, '$.type[0]') != 'card'""",
        order="""json_extract(resources.resource, '$.published[0]') DESC""",
    ):
        r = post["resource"]
        if "entry" in r["type"]:
            r["author"] = tx.identities.get_identity(r["author"][0])["card"]
        yield r


@posts_model.control
def get_posts_by_channel(db, uid):
    """."""
    return db.select(
        "resources",
        vals=[uid],
        where="""json_extract(resources.resource,
                                     '$.channel[0]') == ?""",
        order="""json_extract(resources.resource,
                                     '$.published[0]') DESC""",
    )


# def get_channels(db):
#     """Return a list of channels."""
#     return [r["value"] for r in
#             db.select("""resources,
#                            json_tree(resources.resource, '$.channel')""",
#                          what="DISTINCT value", where="type = 'text'")]


@posts_model.control
def get_year(db, year):
    return db.select(
        "resources",
        order="""json_extract(resources.resource,
                                     '$.published[0].datetime') ASC""",
        where=f"""json_extract(resources.resource,
                                      '$.published[0].datetime')
                                      LIKE '{year}%'""",
    )


@posts_model.control
def get_media(db):
    """Return a list of media filepaths."""
    try:
        filepaths = list(pathlib.Path(tx.host.name).iterdir())
    except FileNotFoundError:
        filepaths = []
    return filepaths


@posts_model.control
def get_filepath(db, filename):
    """Return a media file's path."""
    return pathlib.Path(tx.host.name) / filename



""""""

import json

from understory import term

import micropub

__all__ = ["main"]

main = term.application("micropub", micropub.__doc__)


@main.register()
class Micropub:
    """A Micropub client."""

    # TODO media upload

    def setup(self, add_arg):
        add_arg("endpoint", help="address of the Micropub endpoint")
        add_arg("--type", default="entry", help="post type")
        add_arg("--token", default=None, help="IndieAuth bearer token")
        add_arg("--channel", nargs="*", help="add to given channel(s)")

    def run(self, stdin, log):
        properties = json.loads(stdin.read())
        try:
            properties["channel"].extend(self.channel)
        except KeyError:
            if self.channel:
                properties["channel"] = self.channel
        location, links = micropub.client.create_post(
            properties, endpoint=self.endpoint, h=self.type, token=self.token
        )
        print("Location:", location)
        if links:
            print("Links:", links)
        return 0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['micropub']

package_data = \
{'': ['*']}

install_requires = \
['nltk>=3.6.5,<4.0.0', 'understory>=0,<1']

entry_points = \
{'console_scripts': ['micropub = micropub.__main__:main']}

setup_kwargs = {
    'name': 'micropub',
    'version': '0.0.5',
    'description': 'A library for writing Micropub servers and clients.',
    'long_description': "# micropub-python\n\nA library for writing [Micropub][0] servers and clients.\n\n> The Micropub protocol is used to create, update and delete posts on\n> one's own domain using third-party clients. Web apps and native apps\n> (e.g., iPhone, Android) can use Micropub to post and edit articles,\n> short notes, comments, likes, photos, events or other kinds of posts\n> on your own website.\n\n[0]: https://micropub.spec.indieweb.org/\n",
    'author': 'Angelo Gladding',
    'author_email': 'self@angelogladding.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



"""
utility functions for breaking down a given block of text
into it's component syntactic parts.

"""

import math
import re
import time

import nltk
from nltk.tokenize import RegexpTokenizer

TOKENIZER = RegexpTokenizer(r"(?u)\W+|\$[\d\.]+|\S+")
SPECIAL_CHARS = [".", ",", "!", "?"]


def get_smog(text):
    if not text:
        return 0
    return round(Readability(text).SMOGIndex())


def get_char_count(words):
    characters = 0
    for word in words:
        characters += len(word)
    return characters


def get_words(text=""):
    words = []
    words = TOKENIZER.tokenize(text)
    filtered_words = []
    for word in words:
        if word in SPECIAL_CHARS or word == " ":
            pass
        else:
            new_word = word.replace(",", "").replace(".", "")
            new_word = new_word.replace("!", "").replace("?", "")
            filtered_words.append(new_word)
    return filtered_words


def get_sentences(text=""):
    try:
        tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")
    except LookupError:
        nltk.download("punkt")
        while True:
            try:
                tokenizer = nltk.data.load("tokenizers/punkt/english.pickle")
                break
            except LookupError:
                time.sleep(1)
    sentences = tokenizer.tokenize(text)
    return sentences


def count_syllables(words):
    syllableCount = 0
    for word in words:
        syllableCount += count(word)
    return syllableCount


# This method must be enhanced. At the moment it only
# considers the number of syllables in a word.
# This often results in that too many complex words are detected.
def count_complex_words(text=""):
    words = get_words(text)
    sentences = get_sentences(text)
    complex_words = 0
    found = False
    cur_word = []

    for word in words:
        cur_word.append(word)
        if count_syllables(cur_word) >= 3:

            # Checking proper nouns. If a word starts with a capital letter
            # and is NOT at the beginning of a sentence we don't add it
            # as a complex word.
            if not (word[0].isupper()):
                complex_words += 1
            else:
                for sentence in sentences:
                    if str(sentence).startswith(word):
                        found = True
                        break
                if found:
                    complex_words += 1
                    found = False

        cur_word.remove(word)
    return complex_words


"""
Fallback syllable counter

This is based on the algorithm in Greg Fast's perl module
Lingua::EN::Syllable.

"""

specialSyllables_en = """tottered 2
chummed 1
peeped 1
moustaches 2
shamefully 3
messieurs 2
satiated 4
sailmaker 4
sheered 1
disinterred 3
propitiatory 6
bepatched 2
particularized 5
caressed 2
trespassed 2
sepulchre 3
flapped 1
hemispheres 3
pencilled 2
motioned 2
poleman 2
slandered 2
sombre 2
etc 4
sidespring 2
mimes 1
effaces 2
mr 2
mrs 2
ms 1
dr 2
st 1
sr 2
jr 2
truckle 2
foamed 1
fringed 2
clattered 2
capered 2
mangroves 2
suavely 2
reclined 2
brutes 1
effaced 2
quivered 2
h'm 1
veriest 3
sententiously 4
deafened 2
manoeuvred 3
unstained 2
gaped 1
stammered 2
shivered 2
discoloured 3
gravesend 2
60 2
lb 1
unexpressed 3
greyish 2
unostentatious 5
"""

fallback_cache = {}

fallback_subsyl = ["cial", "tia", "cius", "cious", "gui", "ion", "iou", "sia$", ".ely$"]

fallback_addsyl = [
    "ia",
    "riet",
    "dien",
    "iu",
    "io",
    "ii",
    "[aeiouy]bl$",
    "mbl$",
    "[aeiou]{3}",
    "^mc",
    "ism$",
    "(.)(?!\\1)([aeiouy])\\2l$",
    "[^l]llien",
    "^coad.",
    "^coag.",
    "^coal.",
    "^coax.",
    "(.)(?!\\1)[gq]ua(.)(?!\\2)[aeiou]",
    "dnt$",
]


# Compile our regular expressions
for i in range(len(fallback_subsyl)):
    fallback_subsyl[i] = re.compile(fallback_subsyl[i])
for i in range(len(fallback_addsyl)):
    fallback_addsyl[i] = re.compile(fallback_addsyl[i])


def _normalize_word(word):
    return word.strip().lower()


# Read our syllable override file and stash that info in the cache
for line in specialSyllables_en.splitlines():
    line = line.strip()
    if line:
        toks = line.split()
        assert len(toks) == 2
        fallback_cache[_normalize_word(toks[0])] = int(toks[1])


def count(word):
    word = _normalize_word(word)
    if not word:
        return 0

    # Check for a cached syllable count
    count = fallback_cache.get(word, -1)
    if count > 0:
        return count

    # Remove final silent "e"
    if word[-1] == "e":
        word = word[:-1]

    # Count vowel groups
    count = 0
    prev_was_vowel = 0
    for c in word:
        is_vowel = c in ("a", "e", "i", "o", "u", "y")
        if is_vowel and not prev_was_vowel:
            count += 1
        prev_was_vowel = is_vowel

    # Add & subtract syllables
    for r in fallback_addsyl:
        if r.search(word):
            count += 1
    for r in fallback_subsyl:
        if r.search(word):
            count -= 1

    # Cache the syllable count
    fallback_cache[word] = count

    return count


class Readability:
    """"""

    def __init__(self, text):
        self.analyze(text)

    def analyze(self, text):
        words = get_words(text)
        char_count = get_char_count(words)
        word_count = len(words)
        sentence_count = len(get_sentences(text))
        syllable_count = count_syllables(words)
        complexwords_count = count_complex_words(text)
        avg_words_p_sentence = word_count / sentence_count
        self.stats = {
            "words": words,
            "char_cnt": float(char_count),
            "word_cnt": float(word_count),
            "sentence_cnt": float(sentence_count),
            "syllable_cnt": float(syllable_count),
            "complex_word_cnt": float(complexwords_count),
            "avg_words_p_sentence": float(avg_words_p_sentence),
        }

    @property
    def metrics(self):
        return {
            metric: getattr(self, metric)()
            for metric in (
                "ARI",
                "FleschReadingEase",
                "FleschKincaidGradeLevel",
                "GunningFogIndex",
                "SMOGIndex",
                "ColemanLiauIndex",
                "LIX",
                "RIX",
            )
        }

    def ARI(self):
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            score = (
                4.71 * (self.stats["char_cnt"] / self.stats["word_cnt"])
                + 0.5 * (self.stats["word_cnt"] / self.stats["sentence_cnt"])
                - 21.43
            )
        return score

    def FleschReadingEase(self):
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            score = (
                206.835
                - (1.015 * self.stats["avg_words_p_sentence"])
                - (84.6 * (self.stats["syllable_cnt"] / self.stats["word_cnt"]))
            )
        return round(score, 4)

    def FleschKincaidGradeLevel(self):
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            score = (
                0.39 * self.stats["avg_words_p_sentence"]
                + 11.8 * (self.stats["syllable_cnt"] / self.stats["word_cnt"])
                - 15.59
            )
        return round(score, 4)

    def GunningFogIndex(self):
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            score = 0.4 * (
                self.stats["avg_words_p_sentence"]
                + (100 * (self.stats["complex_word_cnt"] / self.stats["word_cnt"]))
            )
        return round(score, 4)

    def SMOGIndex(self):
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            score = (
                math.sqrt(
                    self.stats["complex_word_cnt"] * (30 / self.stats["sentence_cnt"])
                )
                + 3
            )
        return score

    def ColemanLiauIndex(self):
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            score = (
                (5.89 * (self.stats["char_cnt"] / self.stats["word_cnt"]))
                - (30 * (self.stats["sentence_cnt"] / self.stats["word_cnt"]))
                - 15.8
            )
        return round(score, 4)

    def LIX(self):
        longwords = 0.0
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            for word in self.stats["words"]:
                if len(word) >= 7:
                    longwords += 1.0
            score = (
                self.stats["word_cnt"] / self.stats["sentence_cnt"]
                + float(100 * longwords) / self.stats["word_cnt"]
            )
        return score

    def RIX(self):
        longwords = 0.0
        score = 0.0
        if self.stats["word_cnt"] > 0.0:
            for word in self.stats["words"]:
                if len(word) >= 7:
                    longwords += 1.0
            score = longwords / self.stats["sentence_cnt"]
        return score
